## @package unogenerator.objects
## @brief Objects for unogenerator package
from .unogenerator import ODT, ODS, ODT_Standard, ODS_Standard
from .commons import __version__, __versiondate__,  __versiondatetime__,  Coord,  Range,  ColorsNamed
